#include <Magick++.h>
#include <iostream>
#include <map>
#include <limits.h>
#include <stdlib.h>
#include <fstream>
#include <libgen.h>
#include <math.h>
#include <vector>
#include <list>
#include <set>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <sstream>
#include <assert.h>

using namespace std;
using namespace Magick;

#define PI 3.1415926
#define BLACK 55000
#define BLACK_EDGES 62000
//#define BLACK 62000

#define line pair<double, double>

#define BOXED 1

struct Stump
{
  unsigned ind;
  double thresh;
  bool low;
  double weight;
};

struct Cluster
{
  unsigned start;
  unsigned end;
  unsigned count;
  unsigned min;
};

vector<Stump> classifier;

vector<double> sinDeg;
vector<double> cosDeg;

map<vector<double>, unsigned> patterns;

#define GRIDSIZE  10
#define GRIDSIZEF 10.0
#define GRIDSIZE2 100

unsigned circle[] = {0, 0, 0, 1, 1, 1, 1, 0, 0, 0,
		     0, 1, 1, 0, 0, 0, 0, 1, 1, 0,
		     0, 1, 0, 0, 0, 0, 0, 0, 1, 0,
		     1, 0, 0, 0, 0, 0, 0, 0, 0, 1,
		     1, 0, 0, 0, 0, 0, 0, 0, 0, 1,
		     1, 0, 0, 0, 0, 0, 0, 0, 0, 1,
		     1, 0, 0, 0, 0, 0, 0, 0, 0, 1,
		     0, 1, 1, 0, 0, 0, 0, 0, 1, 0,
		     0, 0, 1, 0, 0, 0, 0, 1, 1, 0,
		     0, 0, 0, 1, 1, 1, 1, 0, 0, 0};

unsigned filled[] = {0, 0, 0, 1, 1, 1, 1, 1, 0, 0,
		     0, 1, 1, 1, 1, 1, 1, 1, 1, 0,
		     0, 1, 1, 1, 1, 1, 1, 1, 1, 0,
		     1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		     1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		     1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		     1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		     0, 1, 1, 1, 1, 1, 1, 1, 1, 0,
		     0, 0, 1, 1, 1, 1, 1, 1, 1, 0,
		     0, 0, 0, 1, 1, 1, 1, 1, 0, 0};
		     
unsigned crossed[] = {1, 0, 0, 0, 0, 0, 0, 0, 0, 1,
		      0, 1, 0, 0, 0, 0, 0, 0, 1, 0,
		      0, 0, 1, 1, 1, 1, 1, 1, 0, 0,
		      0, 0, 1, 1, 1, 1, 1, 1, 0, 0,
		      0, 0, 1, 1, 1, 1, 1, 1, 0, 0,
		      0, 0, 1, 1, 1, 1, 1, 1, 0, 0,
		      0, 0, 1, 1, 1, 1, 1, 1, 0, 0,
		      0, 0, 1, 1, 1, 1, 1, 1, 0, 0,
		      0, 1, 0, 0, 0, 0, 0, 0, 1, 0,
		      1, 0, 0, 0, 0, 0, 0, 0, 0, 1};

unsigned boxed[] = {1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		    1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		    1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		    1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		    1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		    1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		    1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		    1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		    1, 1, 1, 1, 1, 1, 1, 1, 1, 1,
		    1, 1, 1, 1, 1, 1, 1, 1, 1, 1};


ostream& operator << (ostream& s, const line& l)
{
  s << "Line t = " << l.first << " r = " << l.second;
  return s;
}

PixelPacket* imCache = NULL;
unsigned colsCache = 0;
unsigned rowsCache = 0;

inline double avgGetPixel(Image& im, unsigned int x, unsigned int y)
{
  if (x >= colsCache) x = colsCache - 1;
  if (y >= rowsCache) y = rowsCache - 1;

  Color c = im.pixelColor(x, y);
 
  return c.redQuantum();
}

inline double dist(line a, line b)
{
  double dt = (a.first - b.first);
  double dr = a.second - b.second;
  if (dr > 180)
    dr -= 360;
  if (dr < -180)
    dr += 360;

  dr *= 10;

  return dt * dt + dr * dr;
}

line closest(vector<line>& v, line l)
{
  assert(v.size() != 0);

  double d = dist(v[0], l);
  unsigned ind = 0;
  for (unsigned i = 1; i < v.size(); i++)
    {
      double t = dist(v[i], l);
      if (t < d)
	{
	  d = t;
	  ind = i;
	}
    }
  line temp = v[ind];
  v.erase(v.begin() + ind);
  return temp;
}

inline bool dominates(vector<vector<unsigned> >& a, unsigned r, unsigned t)
{
  if (r > 0 and a[r - 1][t] >= a[r][t]) return false;
  if (t > 0 and a[r][t - 1] > a[r][t]) return false;
  if (r < a.size() - 1 and a[r + 1][t] > a[r][t]) return false;
  if (t < a[r].size() - 1 and a[r][t + 1] >= a[r][t]) return false;
  return true;
}

void swap(unsigned& a, unsigned& b)
{
  unsigned t = a;
  a = b;
  b = t;
}

Image crop(Image& im, unsigned x1, unsigned y1, unsigned x2, unsigned y2)
{
  if (x2 < x1) swap(x1, x2);
  if (y2 < y1) swap(y1, y2);
  Image ret(Geometry(x2 - x1, y2 - y1), Color("white"));

  for (unsigned y = y1; y < y2; y++)
    {
      for (unsigned x = x1; x < x2; x++)
	{
	  ret.pixelColor(x - x1, y - y1, im.pixelColor(x, y));
	}
    }

  return ret;
}


vector<line> findEdges(Image& im)
{
  int d = (int)ceil(sqrt((double)(im.columns() * im.columns() +
				  im.rows() * im.rows())));
  vector<vector<unsigned> > a(d, vector<unsigned>(360));

  unsigned sx = (unsigned)(im.columns() * 0.01);
  unsigned sy = (unsigned)(im.rows() * 0.01);
  unsigned mx = (unsigned)(im.columns() * 0.98);
  unsigned my = (unsigned)(im.rows() * 0.98);

  colsCache = im.columns();
  rowsCache = im.rows();
  imCache = im.getPixels(0, 0, colsCache, rowsCache);


  set<line> possibles;

  const double POSSIBLE_THRESH = 0.22;
  const unsigned THRESH = (int)(d * POSSIBLE_THRESH);

  for (unsigned x = sx; x < mx; x++)
    {
      for (unsigned y = sy + (x & 1); y < my; y+=2)
	{
	  if (avgGetPixel(im, x, y) < BLACK_EDGES)
	    {
	      for (unsigned t = 0; t < 5; t++)
		{
		  int r = (int)floor(x * sinDeg[t] + 
				     y * cosDeg[t]);
		  if (r > 0 and r < d)
		    {
		      a[r][t] += 1;
		      if (a[r][t] == THRESH)
			{
			  possibles.insert(line(r, t));
			}
		    }
		}
	      for (unsigned t = 85; t < 95; t++)
		{
		  int r = (int)floor(x * sinDeg[t] + 
				     y * cosDeg[t]);
		  if (r > 0 and r < d)
		    {
		      a[r][t] += 1;
		      if (a[r][t] == THRESH)
			{
			  possibles.insert(line(r, t));
			}
		    }
		}
	      for (unsigned t = 175; t < 185; t++)
		{
		  int r = (int)floor(x * sinDeg[t] + 
				     y * cosDeg[t]);
		  if (r > 0 and r < d)
		    {
		      a[r][t] += 1;
		      if (a[r][t] == THRESH)
			{
			  possibles.insert(line(r, t));
			}
		    }
		}
	      for (unsigned t = 265; t < 275; t++)
		{
		  int r = (int)floor(x * sinDeg[t] + 
				     y * cosDeg[t]);
		  if (r > 0 and r < d)
		    {
		      a[r][t] += 1;
		      if (a[r][t] == THRESH)
			{
			  possibles.insert(line(r, t));
			}
		    }
		}
	      for (unsigned t = 355; t < 360; t++)
		{
		  int r = (int)floor(x * sinDeg[t] + 
				     y * cosDeg[t]);
		  if (r > 0 and r < (int)d)
		    {
		      a[r][t] += 1;
		      if (a[r][t] == THRESH)
			{
			  possibles.insert(line(r, t));
			}
		    }
		}
	    }
	}
    }

  vector<line > good;
  for (set<line>::iterator i = possibles.begin(); i != possibles.end(); ++i)
    {
      if (dominates(a, (unsigned)i->first, (unsigned)i->second))
	{
	  good.push_back(*i);
	}
    }
  
  vector<line> ret;
  // Left
  ret.push_back(closest(good, line(im.columns() * 0.08, 90)));
  // Right
  ret.push_back(closest(good, line(im.columns() * 0.92, 90)));
  // Top 
  ret.push_back(closest(good, line(im.rows() * 0.08, 0)));
  // Bottom
  ret.push_back(closest(good, line(im.rows() * 0.85, 0)));


  return ret;
}

void calcTrig()
{
  for (unsigned i = 0; i < 360; i++)
    {
      sinDeg.push_back(sin(i * PI/180));
      cosDeg.push_back(cos(i * PI/180));
    }
}

void initImage(Image& im, Image& tempIm)
{
  calcTrig();

  // Flood fill corners to deal with bad staple pulls and such
  im.colorFuzz(25000);
  if (avgGetPixel(im, 0, 0) < BLACK)
    im.floodFillColor(0, 0, Color("White"));
  if (avgGetPixel(im, 0, im.rows() - 1) < BLACK)
    im.floodFillColor(0, im.rows() - 1, Color("White"));
  if (avgGetPixel(im, im.columns() - 1, 0) < BLACK)
    im.floodFillColor(im.columns() - 1, 0, Color("White"));
  if (avgGetPixel(im, im.columns() - 1, im.rows() - 1) < BLACK)
    im.floodFillColor(im.columns() - 1, im.rows() - 1, Color("White"));

  // Flood fill corners to deal with bad staple pulls and such
  tempIm.colorFuzz(25000);
  if (avgGetPixel(tempIm, 0, 0) < BLACK)
    tempIm.floodFillColor(0, 0, Color("White"));
  if (avgGetPixel(tempIm, 0, tempIm.rows() - 1) < BLACK)
    tempIm.floodFillColor(0, tempIm.rows() - 1, Color("White"));
  if (avgGetPixel(tempIm, tempIm.columns() - 1, 0) < BLACK)
    tempIm.floodFillColor(tempIm.columns() - 1, 0, Color("White"));
  if (avgGetPixel(tempIm, tempIm.columns() - 1, tempIm.rows() - 1) < BLACK)
    tempIm.floodFillColor(tempIm.columns() - 1, tempIm.rows() - 1, 
			  Color("White"));

  tempIm.edge(1);
  tempIm.negate(true);
  tempIm.blur(2);

  vector<line> edges = findEdges(tempIm);
  line leftEdge = edges[0];
  line rightEdge = edges[1];
  line topEdge = edges[2];
  line bottomEdge = edges[3];
  
  double rotation = (leftEdge.second - 90 + rightEdge.second - 90) / 2.0;
  
  if (rotation > 90) 
    rotation -= 180;
  else if (rotation < -90)
    rotation += 180;  

  //cerr << "Rotation is " << rotation << endl;
  if (rotation != 0)
    im.rotate(rotation);
  
  // After rotating, the vertical lines update only r, the horizontals
  // update both
  leftEdge.second -= rotation;
  rightEdge.second -= rotation;
  
  if (rotation < 0)
    {
      topEdge.first -= ((int)(sin(rotation * PI / 180) * im.columns()));
      bottomEdge.first -= ((int)(sin(rotation * PI / 180) * im.columns()));
    } else {
      topEdge.first += ((int)(sin(rotation * PI / 180) * im.columns()));
      bottomEdge.first += ((int)(sin(rotation * PI / 180) * im.columns()));
    }

  topEdge.second -= rotation;
  bottomEdge.second -= rotation;

  if (rotation > 0)
    {
      rightEdge.first += (int)(sin(rotation * PI / 180) * im.rows());
      leftEdge.first += (int)(sin(rotation * PI / 180) * im.rows());      
    }
  
  double skew = tan((topEdge.second + bottomEdge.second) / 2.0 * PI / 180.0);
  // TODO: is abs right here?
  //int dy = (int)abs(sin((topEdge.second + bottomEdge.second)/ 2.0 * 
  //			PI / 180.0) * im.columns() * 1.5);

  if (skew > -1e-4 and skew < 1e-4) skew = 0;
  
  if (skew < 0)
    {
      im.flop();
      DrawableAffine affine(1.0, 1.0, skew*-1, 0.0, 0.0, 0);
      im.affineTransform(affine);
      im.flop();
    }
  else if (skew > 0)
    {
      DrawableAffine affine(1.0, 1.0, skew, 0.0, 0.0, 0);
      im.affineTransform(affine);      
    }
  
  //cerr << "Dy is " << dy << endl;  
  //topEdge.first += dy;
  //bottomEdge.first += dy;
  
  im = crop(im, 
	    (unsigned)leftEdge.first, (unsigned)topEdge.first, 
	    (unsigned)rightEdge.first, (unsigned)bottomEdge.first);

  im.blur(2);
  im.type(GrayscaleType);
  im.quantizeColors(2);
  
  //im.write("crop.png");
  
}

template <typename T>
inline T min(const vector<T>& v)
{
  T ret = v[0];
  for (unsigned i = 1; i < v.size(); i++)
    {
      if (v[i] < ret) ret = v[i];
    }
  return ret;
}

template <typename T>
inline T max(const vector<T>& v)
{
  T ret = v[0];
  for (unsigned i = 1; i < v.size(); i++)
    {
      if (v[i] > ret) ret = v[i];
    }
  return ret;
}

inline int min(double a, int b)
{
  if (a < b) return (int)a;
  return b;
}

inline int max(double a, int b)
{
  if (a > b) return (int)a;
  return b;
}


inline double average(const vector<double>& v)
{
  if (v.size() == 0) return 0.0;
  double sum = 0;
  for (unsigned int i = 0; i < v.size(); ++i)
    {
      sum += v[i];
    }
  return (sum) / v.size();
}

inline double dist(const vector<double>& v1, 
	    const vector<double>& v2)
{
  assert(v1.size() == v2.size());
  double ret = 0.0;
  for (unsigned i = 0; i < v1.size(); i++)
    {
      ret += (v1[i] - v2[i]) * (v1[i] - v2[i]);
    }
  return ret;
}

double classify(const vector<double>& v)
{
  double vote = 0.0;
  int mult;
  for (unsigned i = 0; i < classifier.size(); i++)
    {
      if (classifier[i].low) 
	mult = 1;
      else
	mult = -1;
      
      if (v[classifier[i].ind] <= classifier[i].thresh)
	vote += mult * classifier[i].weight;
      else
	vote -= mult * classifier[i].weight;
    }
  //cerr << "Vote is " << vote << endl;
  return vote;
}

vector<double> sample(Image& im,
		      unsigned cx, 
		      unsigned cy)
{
  unsigned radius = (unsigned)(0.01 * im.columns());

  vector<double> xs, ys;
  list<pair<unsigned, unsigned> > toCheck;
  set<pair<unsigned, unsigned> > seen;

  // Move outward until we find some filled pixels.
  bool left = true;
  bool right = true;
  bool up = true;
  bool down = true;
  for (unsigned d = 1; d < radius and toCheck.size() < 4; d++)
    {
      if (right and avgGetPixel(im, cx + d, cy) < BLACK)
	{
	  pair<unsigned, unsigned> temp(cx + d, cy);
	  seen.insert(temp);
	  toCheck.push_back(temp);
	  xs.push_back(cx + d);
	  ys.push_back(cy);
	  right = false;
	}
      if (left and avgGetPixel(im, cx - d, cy) < BLACK)
	{
	  pair<unsigned, unsigned> temp(cx - d, cy);
	  seen.insert(temp);
	  toCheck.push_back(temp);
	  xs.push_back(cx - d);
	  ys.push_back(cy);
	  left = false;
	}
      if (down and avgGetPixel(im, cx, cy + d) < BLACK)
	{
	  pair<unsigned, unsigned> temp(cx, cy + d);
	  seen.insert(temp);
	  toCheck.push_back(temp);
	  xs.push_back(cx - d);
	  ys.push_back(cy + d);
	  down = false;

	}
      if (up and avgGetPixel(im, cx, cy - d) < BLACK)
	{
	  pair<unsigned, unsigned> temp(cx, cy - d);
	  seen.insert(temp);
	  toCheck.push_back(temp);
	  xs.push_back(cx);
	  ys.push_back(cy - d);
	  up = false;
	}
    }
  
  if (toCheck.empty()) return vector<double>();

  // Now flood-fill from this point.
  while (not toCheck.empty())
    {
      pair<unsigned, unsigned> cur = toCheck.front();
      toCheck.pop_front();
      unsigned curX = cur.first;
      unsigned curY = cur.second;
      if (abs((int)(curX - cx)) > 30 or abs((int)(curY - cy)) > 30) continue;

      for (int dx = -1; dx <= 1; dx++)
	{
	  for (int dy = -1; dy <= 1; dy++)
	    {
	      pair<unsigned, unsigned> temp(curX + dx, curY + dy);
	      if (avgGetPixel(im, curX + dx, curY + dy) < BLACK and 
		  seen.count(temp) == 0)
		{
		  seen.insert(temp);
		  toCheck.push_back(temp);
		  xs.push_back(curX + dx);
		  ys.push_back(curY + dy);
		}
	    }
	}
    }

  //cerr << "Touched " << xs.size() << " points" << endl;

  unsigned width = (unsigned)(max(xs) - min(xs));
  unsigned height = (unsigned)(max(ys) - min(ys));
  
  cx = (unsigned)average(xs) - width / 2;
  cy = (unsigned)average(ys) - height / 2;

  vector<double> ret;
  for (unsigned i = 0; i < GRIDSIZE; i++)
    {
      for (unsigned j = 0; j < GRIDSIZE; j++)
	{
	  ret.push_back(avgGetPixel(im, 
				    (unsigned)(j * (width/(GRIDSIZEF-1)) + cx),
				    (unsigned)(i *(height/(GRIDSIZEF-1))+cy)));
	  //continue;
	  unsigned tx = (unsigned)(j * (width/(GRIDSIZEF-1)) + cx);
	  unsigned ty = (unsigned)(i * (height/(GRIDSIZEF-1)) + cy);
	  if (tx >= im.columns()) tx = im.columns() - 1;
	  if (ty >= im.rows()) ty = im.rows() - 1;

	  im.pixelColor(tx, ty, Color("blue"));
	}
    }

  vector<double> best;
  double bestD = -1.0;
  for (map<vector<double>, unsigned>::iterator i = patterns.begin(); 
       i != patterns.end(); ++i)
    {
      double d = dist(i->first, ret);
      if (d < bestD or bestD == -1.0)
	{
	  best = i->first;
	  bestD = d;
	}
    }
  
  ret.push_back(average(ret));
  ret.push_back(width);
  ret.push_back(height);  
  ret.push_back(width*height);
  ret.push_back((double)patterns[best]);
  return ret;

}


template<typename T>
void printVec(const vector<T>& v)
{
  for (unsigned i = 0; i < v.size(); i++)
    {
      cout << v[i] << " ";
    }
  cout << endl;
}

template<typename T>
T sum(const vector<T>& v)
{
  T ret = v[0];
  for (unsigned i = 1; i < v.size(); i++)
    ret += v[i];
  return ret;
}


vector<unsigned> aggCluster(vector<unsigned>& v, unsigned k)
{
  list<Cluster> cs;
  map<unsigned, unsigned> counts;
  unsigned maxCount = 0;
  for (unsigned i = 0; i < v.size(); i++)
    {
      counts[v[i]]++;
      if (counts[v[i]] > maxCount) maxCount++;
    }

  for (map<unsigned, unsigned>::iterator i = counts.begin(); 
       i != counts.end(); ++i)
    {
      Cluster c;
      c.start = i->first;
      c.end = i->first;
      c.count = i->second;
      c.min = i->second;
      if (i->second * 10 > maxCount)
	cs.push_back(c);     
    }

  list<Cluster>::iterator best;
  list<Cluster>::iterator next;
  list<Cluster>::iterator i, j;
  unsigned bestDist;

  while (cs.size() > k) 
    {
      // Find the best two adjacent Clusters;
      best = cs.begin();
      next = cs.begin();
      ++next;

      bestDist = 0;
      for (i = best, j = next;
	   j != cs.end(); 
	   ++i, ++j)
	{	  
	  unsigned dist = (i->count + j->count)/(j->end - i->start);
	  if (dist > bestDist) {
	    best = i;
	    next = j;
	    bestDist = dist;
	  }
	}

      // Merge 'em
      best->end = next->end;
      best->count += next->count;
      best->min = bestDist;

      cs.erase(next);
    }

  // Now for each cluster, output the weighted center of it
  vector<unsigned> ret;
  ret.reserve(k);

  for (list<Cluster>::iterator i = cs.begin(); i != cs.end(); ++i) 
    {
      unsigned sum = 0;
      unsigned count = 0;
      for (unsigned j = i->start; j <= i->end; j++) {
	sum += counts[j] * j;
	count += counts[j];
      }

      ret.push_back(sum/count);
    }
  return ret;

}

vector<unsigned> cluster(vector<unsigned>& v, unsigned k)
{
  sort(v.begin(), v.end());

  unsigned minVal = v[0];
  unsigned maxVal = v[v.size() - 1];
  
  vector<unsigned> clusters(k);
  
  vector<unsigned> assignments(v.size());

  const unsigned NUM_CLUSTER_ROUNDS = 3 * k * k + 10;
  unsigned bestError = sum(v);
  vector<unsigned> ret;

  if (maxVal - minVal == 0) return clusters;

  vector<unsigned> sums(k);
  vector<unsigned> counts(k);

  unsigned vs = v.size();
  unsigned moved = true;
  for (unsigned round = 0; round < NUM_CLUSTER_ROUNDS; round++)
    {
      // Initialize 'em
      for (unsigned i = 0; i < k; i++)
	{
	  if (round == 0)
	    clusters[i] = i * (maxVal - minVal) / k;
	  else
	    clusters[i] = rand() % (maxVal - minVal) + minVal;
	}

      // These shall always be sorted
      if (round != 0)
	sort(clusters.begin(), clusters.end());

      moved = true;
      while (1)
	{      
	  // Assign to nearest clusters
	  unsigned cluster = 0;
	  unsigned nextSplit = (clusters[0] + clusters[1]) / 2;

	  for (unsigned i = 0; i < k; i++) {
	    sums[i] = 0;
	    counts[i] = 0;
	  }

	  for (unsigned i = 0; i < vs; i++)
	    {
	      if (v[i] > nextSplit) {
		cluster++;
		if (cluster + 1 < k)
		  nextSplit = (clusters[cluster] + clusters[cluster + 1]) / 2;
		else
		  nextSplit = maxVal + 1;
	      }

	      if (assignments[i] != cluster) 
		{
		  moved = true;
		  assignments[i] = cluster;
		}

	      sums[assignments[i]] += v[i];
	      counts[assignments[i]]++;
	    }

	  //cout << "Moved: " << moved << " of " << v.size() << endl;
	  if (not moved) break;

	  // Now recalculate centers
	  bool needsSort = false;
	  for (unsigned i = 0; i < k; i++) 
	    {
	      if (counts[i]) {
		clusters[i] = sums[i] / counts[i];
	      } else {
		clusters[i] = rand() % (maxVal - minVal) + minVal;
		needsSort = true;
	      }
	    }
	  
	  if (needsSort) 
	    sort(clusters.begin(), clusters.end());	 

	  moved = false;
	}    
      
      unsigned error = 0;
      for (unsigned i = 0; i < vs; i++)
	{
	  error += abs((int)v[i] - (int)clusters[assignments[i]]);
	}

      if (round == 0 or error < bestError)
	{
	  ret = clusters;
	  bestError = error;
	}
    }

  sort(ret.begin(), ret.end());
  
  return ret;
}


vector<pair<unsigned, unsigned> > getCenters(Image& im,
					     double pSX, double pSY,
					     double width, double height,
					     unsigned numRows,
					     unsigned numCols)
{
  vector<unsigned> xs;
  vector<unsigned> ys;
  
  unsigned sx = (unsigned)(pSX * im.columns());
  unsigned sy = (unsigned)(pSY * im.rows());

  unsigned mx = sx + (unsigned)(width * im.columns());
  unsigned my = sy + (unsigned)(height * im.rows());
  
  for (unsigned y = sy; y < my; y++)
    {
      for (unsigned x = sx + (y & 1); x < mx; x += 2)
	{
	  if (avgGetPixel(im, x, y) < BLACK)
	    {
	      xs.push_back(x);
	      ys.push_back(y);
	    }
	}
    }

  xs = cluster(xs, numCols);
  ys = cluster(ys, numRows);

  vector<pair<unsigned, unsigned> > ret(ys.size() * xs.size());
  int cur = 0;
  for (unsigned y = 0; y < ys.size(); y++)        
   {
     for (unsigned x = 0; x < xs.size(); x++)
	{
	  ret[cur++] = pair<unsigned, unsigned>(xs[x], ys[y]);	  
	}
    }

#if BOXED
  list<Drawable> drawList;
  drawList.push_back(DrawableStrokeColor("red"));
  drawList.push_back(DrawableStrokeWidth(1));
  for (unsigned x = 0; x < xs.size(); x++)
    {
      drawList.push_back(DrawableLine(xs[x], ys[0], xs[x], ys[ys.size() - 1]));
    }

  for (unsigned y = 0; y < ys.size(); y++)
    {
      drawList.push_back(DrawableLine(xs[0], ys[y], xs[xs.size() - 1], ys[y]));
    }

  //im.draw(drawList);
  //im.write("clustered.png");  
#endif

  //cerr << "Ret size is " << ret.size() << endl;

  return ret;
}

vector<vector<double> > extractGrid(Image& im,
				    unsigned numRows,
				    unsigned numCols,
				    double pSX,
				    double pSY,
				    double width,
				    double height,
				    bool training)
  
{				    
  list<Drawable> drawList;

  vector<pair<unsigned, unsigned> > centers = getCenters(im, 
							 pSX, pSY, 
							 width, height,
							 numRows, numCols);

  unsigned y;
  unsigned x;

  unsigned row = 0;
  unsigned col = 0;

  // Now for each box, sample the bubble image.  Feed this into our
  // classifier to determine filled/empty.
  vector<vector<vector<double> > > samples;
  samples.resize(numRows);
  vector<vector<double> > ret;
  ret.resize(numRows);
  double minAvg = -1;
  double maxAvg = -1;
  for (row = 0; row < numRows; row++)
    {
      for (col = 0; col < numCols; col++)
	{	  
	  //cerr << row << " x " << col;
	  x = centers[row * numCols + col].first;
	  y = centers[row * numCols + col].second;
	  //cerr << ": " << x << ", " << y << endl;

	  vector<double> v = sample(im, x, y);
	  samples[row].push_back(v);

	  if (v.empty()) continue;

	  if (minAvg == -1 or v[GRIDSIZE2] < minAvg) minAvg = v[GRIDSIZE2];
	  if (maxAvg == -1 or v[GRIDSIZE2] > maxAvg) maxAvg = v[GRIDSIZE2];
	}
    }

  double avgRange = maxAvg - minAvg;
  for (row = 0; row < numRows; row++)
    {
      for (col = 0; col < numCols; col++)
	{	  
	  vector<double> v = samples[row][col];
	  // Figure out where we are on the 0 - 1 range for darkness.
	  if (not v.empty()) {
	    double context = (v[GRIDSIZE2] - minAvg)/avgRange;
	    v.push_back(context);
	    if (training) printVec(v);
	    ret[row].push_back(classify(v));
	  } else {
	    ret[row].push_back(-1000);
	  }
	}
    }

#ifdef BOXED
  //im.write("boxed.png");
#endif

  return ret;
}


void usage()
{
  cout << "Usage: homr configFile numBubbles imageFile [-o]" << endl;
  cout << endl;
  cout << "configFile is the path to a file describing layout of" << endl;
  cout << "bubble grids within the page bounding box.  Each grid" << endl;
  cout << "is represented by 7 entries:" << endl;
  cout << endl;
  cout << "Rows Cols SX SY Width Height Order" << endl;
  cout << endl;
  cout << "Where:" << endl;
  cout << " Rows is integer number of rows in grid" << endl;
  cout << " Cols is integer number of columns in grid" << endl;
  cout << " SX is a float starting x position, in range [0 .. 1]" << endl;
  cout << "    with 0 as the left margin, and 1 as the right." << endl;
  cout << " SY is a float starting y position, in range [0 .. 1]" << endl;
  cout << "    with 0 as the top margin, and 1 as the bottom." << endl;
  cout << " Width is the width of the grid, in the range [0 .. 1]" << endl;
  cout << "    as a percentage of the width of the bounding box." << endl;
  cout << " Height is the height of the grid, in the range [0 .. 1]" << endl;
  cout << "    as a percentage of the height of the bounding box." << endl;
  cout << " Order is either D or R, for Down or Right, determining" << endl;
  cout << "    whether bubbles are processed rows or columns first." << endl;
  cout << " It is generally suggested that grids be no larger than " << endl;
  cout << "  15-20 in any dimension.  Larger grids can and should " << endl;
  cout << "  be subdivided." << endl;
  cout << "An example grid:" << endl;
  cout << " 10 9 0.280 0.341 0.570 0.220 D" << endl;
  cout << "This is a 10 rows x 9 cols grid starting at (.28, .341)" << endl;
  cout << "with a width of .57 of the bounding box, and height .22" << endl;
  cout << "Bubbles are read out by columns (D, rather than R)" << endl;
  cout << endl;
  cout << "numBubbles is how many bubbles should be extracted from" << endl;
  cout << " the given page.  It must be less than the number of" << endl;
  cout << " bubbles defined in the given config file." << endl;
  cout << endl;
  cout << "imageFile is the image to process.  It MUST have all four" << endl;
  cout << " borders intact, and must be layed out according to the" << endl;
  cout << " given configuration file. " << endl;
  cout << " Suggested resolution is 150 dpi" << endl << endl;

  cout << "-o implies that there should be only One bubble per row or column."
       << endl;
  exit(1);
}

void setPlatonics()
{
  vector<double> v;
  for (unsigned i = 0; i < GRIDSIZE2; i++)
    {
      if (circle[i])
	v.push_back(0);
      else
	v.push_back(0xFFFF);
    }
  assert(v.size() == GRIDSIZE2);
  patterns[v] = 0;

  for (unsigned i = 0; i < GRIDSIZE2; i++)
    {
      if (crossed[i])
	v[i] = 0;
      else
	v[i] = 0xFFFF;
    }
  assert(v.size() == GRIDSIZE2);
  patterns[v] = 1;
  
  for (unsigned i = 0; i < GRIDSIZE2; i++)
    {
      if (filled[i])
	v[i] = 0;
      else
	v[i] = 0xFFFF;
    }
  assert(v.size() == GRIDSIZE2);
  patterns[v] = 2;

  
  for (unsigned i = 0; i < GRIDSIZE2; i++)
    {
      v[i] = 0;
    }
  patterns[v] = 2;
}

int main(int argc, char** argv)
{
  if (argc < 4)
    usage();

  string configFile(argv[1]);
  unsigned numBubbles = atoi(argv[2]);
  string filename(argv[3]);

  map<string, string> mapping;
  string sid;
  string email;
  char exeLoc[32];
  bool one = false;

  if (argc == 5 and (string)argv[4] == "-o") {
    one = true;
  }

  sprintf(exeLoc, "/proc/%d/exe", getpid());
  argv[0] = exeLoc;
  argv[0] = realpath(argv[0], new char[PATH_MAX]);


  char* classFile = new char[strlen(argv[0]) + strlen("/classify.dat")];
  strcpy(classFile, argv[0]);
  classFile = dirname(classFile);
  strcat(classFile, "/classify.dat");

  ifstream classStream(classFile);
  unsigned ind;
  double thresh;
  char low;
  double weight;
  while (classStream >> ind >> thresh >> low >> weight)
    {
      Stump temp;
      temp.ind = ind;
      temp.thresh = thresh;
      temp.low = (low == 'T');
      temp.weight = weight;
      classifier.push_back(temp);
    }

  setPlatonics();

  Image im;

  try 
    {
      im.read(filename.c_str());
    } 
  catch (Exception& e)
    {
      cerr << "Unable to open " << filename << endl;
      return 1;
    }

  ifstream config(configFile.c_str());
  unsigned curBubble = 0;
  if (not config)
    {
      cerr << "Unable to open " << configFile << endl;
      return 1;
    }

  Image im2 = im;
  im.modifyImage();

  colsCache = im.columns();
  rowsCache = im.rows();
  imCache = im.getPixels(0, 0, colsCache, rowsCache);

  initImage(im, im2);

  colsCache = im.columns();
  rowsCache = im.rows();
  imCache = im.getPixels(0, 0, colsCache, rowsCache);

  bool training = (argc == 5 and not one);
  while (curBubble < numBubbles and config)
    {
      unsigned rows, cols;
      double psx, psy, pwidth, pheight;
      char direction;

      config >> rows >> cols >> psx >> psy >> pwidth >> pheight >> direction;

      vector<vector<double> > grid = extractGrid(im, 
						 rows,
						 cols, 
						 psx,
						 psy,
						 pwidth,
						 pheight,
						 training);
      if (training)
	{
	  curBubble += rows * cols;
	  continue;
	}
      
      if (not one) 
	{
	  if (direction == 'D')
	    {
	      for (unsigned c = 0; c < cols; c++)
		{
		  for (unsigned r = 0; r < rows; r++)
		    {
		      if (grid[r][c] >= 0.0) 
			cout << "1";
		      else
			cout << "0";

		      curBubble++;
		      if (curBubble == numBubbles) 
			{
			  cout << endl;
			  return 0;
			}
		    }
		}	    
	    }
	  else
	    {
	      for (unsigned r = 0; r < rows; r++)
		{
		  for (unsigned c = 0; c < cols; c++)	      
		    {
		      if (grid[r][c] >= 0.0) 
			cout << "1";
		      else
			cout << "0";

		      curBubble++;
		      if (curBubble == numBubbles) 
			{
			  cout << endl;
			  return 0;
			}
		    }
		}	    	  
	    }	 
	} else {
	  // Only one per row/col
	  if (direction == 'D') 
	    {
	      for (unsigned c = 0; c < cols; c++) 
		{
		  unsigned bestInd = 0;
		  for (unsigned r = 0; r < rows; r++) 
		    {
		      if (grid[r][c] > grid[bestInd][c]) 
			bestInd = r;
		    } 
		  for (unsigned r = 0; r < rows; r++)
		    {
		      if (r != bestInd) 
			cout << "0";
		      else 
			cout << "1";
		      curBubble++;
		      if (curBubble == numBubbles)
			{
			  cout << endl;
			  return 0;
			}
		    }
		}
	    } else {
	      for (unsigned r = 0; r < rows; r++) 
		{
		  unsigned bestInd = 0;
		  for (unsigned c = 0; c < cols; c++) 
		    {
		      if (grid[r][c] > grid[r][bestInd]) 
			bestInd = c;
		    } 
		  for (unsigned c = 0; c < cols; c++)
		    {
		      if (c != bestInd) 
			cout << "0";
		      else 
			cout << "1";
		      curBubble++;
		      if (curBubble == numBubbles)
			{
			  cout << endl;
			  return 0;
			}
		    }
		}
	    }
	}
    }
   
  if (not config)
    {
      cerr << "Error in config file: not enough bubble data for " 
	   << numBubbles << endl; 
      return 1;
    }

}
